import logging
import d3m.runtime
import d3m.metadata.base
from sqlalchemy.orm import joinedload
from d3m.container import Dataset
from d3m.metadata import base as metadata_base
from alphad3m.schema import database, convert
from multiprocessing import Manager, Process


logger = logging.getLogger(__name__)

@database.with_db
def execute(pipeline_id, dataset, problem, results_path, msg_queue, db):
    # Get pipeline from database

    pipeline = (
        db.query(database.Pipeline)
            .filter(database.Pipeline.id == pipeline_id)
            .options(joinedload(database.Pipeline.modules),
                     joinedload(database.Pipeline.connections))
    ).one()

    logger.info('About to execute pipeline, id=%s, dataset=%r',
                pipeline_id, dataset)

    # Load data
    dataset = Dataset.load(dataset)
    logger.info('Loaded dataset')

    json_pipeline = convert.to_d3m_json(pipeline)
    logger.info('Pipeline to be executed:\n%s',
                '\n'.join([x['primitive']['python_path'] for x in json_pipeline['steps']]))

    d3m_pipeline = d3m.metadata.pipeline.Pipeline.from_json_structure(json_pipeline, )

    runtime = d3m.runtime.Runtime(pipeline=d3m_pipeline, problem_description=problem,
                                  context=metadata_base.Context.TESTING)

    manager = Manager()
    return_dict = manager.dict()
    p = Process(target=worker, args=(runtime, dataset, return_dict))
    p.start()
    p.join(180)  # Maximum 3 minutes
    fit_results = return_dict['fit_results']
    fit_results.check_success()

    if results_path is not None:
        logger.info('Storing fit results at %s', results_path)
        fit_results.values['outputs.0'].to_csv(results_path)
    else:
        logger.info('NOT storing fit results')

    return fit_results.values


def worker(runtime, dataset, return_dict):
    return_dict['fit_results'] = runtime.fit(inputs=[dataset])



from sqlalchemy import Column
from sqlalchemy.types import TypeDecorator, CHAR
from sqlalchemy.dialects.postgresql import UUID as PG_UUID
import uuid


# From http://docs.sqlalchemy.org/en/latest/core/custom_types.html
# #backend-agnostic-guid-type

class UUID(TypeDecorator):
    """Platform-independent UUID type.

    Uses PostgreSQL's UUID type, otherwise uses
    CHAR(32), storing as stringified hex values.

    """
    impl = CHAR

    def load_dialect_impl(self, dialect):
        if dialect.name == 'postgresql':
            return dialect.type_descriptor(PG_UUID())
        else:
            return dialect.type_descriptor(CHAR(32))

    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        elif dialect.name == 'postgresql':
            return str(value)
        else:
            if not isinstance(value, uuid.UUID):
                return "%.32x" % uuid.UUID(value).int
            else:
                # hexstring
                return "%.32x" % value.int

    def process_result_value(self, value, dialect):
        if value is None:
            return value
        else:
            if not isinstance(value, uuid.UUID):
                value = uuid.UUID(value)
            return value


class UuidMixin(object):
    """A mixin that adds a UUID column and sets it at construction time.

    Using the ``default=`` keyword on the Column would only set the attribute
    at insert time.
    """
    id = Column(UUID, primary_key=True, default=uuid.uuid4)

    def __init__(self, **kwargs):
        super(UuidMixin, self).__init__(**kwargs)
        if 'id' not in kwargs:
            self.id = uuid.uuid4()



import numpy as np
import copy
import logging

logger = logging.getLogger(__name__)

NUM_IT = 5
import logging

logger = logging.getLogger(__name__)

class Arena():
    """
    An Arena class where any 2 agents can be pit against each other.
    """
    def __init__(self, player1, player2, game, display=None, logfile=None):
        """
        Input:
            player 1,2: two functions that takes board as input, return action
            game: Game object
            display: a function that takes board as input and prints it (e.g.
                     display in othello/OthelloGame). Is necessary for verbose
                     mode.

        see othello/OthelloPlayers.py for an example. See pit.py for pitting
        human players/other baselines with each other.
        """
        self.player1 = player1
        self.player2 = player2
        self.game = game
        self.display = display
        if not logfile is None:
            self.f = open(logfile, 'a')

    def playGame(self, verbose=False):
        """
        Executes one episode of a game.

        Returns:
            winner: player who won the game (1 if player1, -1 if player2)
        """
        players = [self.player2, None, self.player1]
        curPlayer = 1
        board = self.game.getInitBoard()
        it = 0
        while self.game.getGameEnded(board, curPlayer)==0 and it <= NUM_IT:
            it+=1
            if verbose:
                self.f.write(','.join(self.game.get_pipeline_primitives(board))+'\n')
                assert(self.display)
                logger.info("Turn %s",it)
                logger.info("Player %s", curPlayer)
                self.display(board)
            action = players[curPlayer+1](self.game.getCanonicalForm(board, curPlayer))
            #print('ACTION ', action)
            valids = self.game.getValidMoves(self.game.getCanonicalForm(board, curPlayer),1)

            #print('VALIDS ', valids)
            if valids[action]!=0:
                #print(action)
                assert valids[action] >0
            board, curPlayer = self.game.getNextState(board, curPlayer, action)
        if verbose:
            self.f.write(','.join(self.game.get_pipeline_primitives(board))+'\n')            
            assert(self.display)
            logger.info("Turn %s", it)
            logger.info("Player %s", curPlayer)
            self.display(board)
        game_ended = self.game.getGameEnded(board, 1)
        if verbose:
            if game_ended == 1:
                self.f.write('Working Pipeline\n')
            elif game_ended == 2:
                self.f.write('Non-Working Pipeline\n')
        #return game_ended==1 and curPlayer == 1
        return self.game.getGameEnded(board, 1)

    def playGames(self, num, verbose=False):
        """
        Plays num games in which player1 starts num/2 games and player2 starts
        num/2 games.

        Returns:
            oneWon: games won by player1
            twoWon: games won by player2
        """
        num = int(num/2)
        oneWon = 0
        twoWon = 0
        if verbose:
            self.f.write('Round 1\n')
        for i in range(num):
            if verbose:
                self.f.write('Game '+str(i)+'\n')
            if self.playGame(verbose=verbose)==1:
                oneWon+=1
            else:
                twoWon+=1
        self.player1, self.player2 = self.player2, self.player1
        if verbose:
            self.f.write('Round 2 - Players Swapped\n')
        for i in range(num):
            if verbose:
                self.f.write('Game '+str(i)+'\n')
            if self.playGame(verbose=verbose)==-1:
                oneWon+=1
            else:
                twoWon+=1
        return oneWon, twoWon



"""Convert to/from the JSON representation.
"""

import importlib
import pickle
import logging
from d3m.utils import compute_digest

logger = logging.getLogger(__name__)


def get_class(name):
    package, classname = name.rsplit('.', 1)
    return getattr(importlib.import_module(package), classname)


def _add_step(steps, modules, params, module_to_step, mod):
    if mod.id in module_to_step:
        return module_to_step[mod.id]

    # Special case: the "dataset" module
    if mod.package == 'data' and mod.name == 'dataset':
        module_to_step[mod.id] = 'inputs.0'
        return 'inputs.0'
    elif mod.package != 'd3m':
        raise ValueError("Got unknown module '%s:%s'", mod.package, mod.name)

    # Recursively walk upstream modules (to get `steps` in topological order)
    # Add inputs to a dictionary, in deterministic order
    inputs = {}

    for conn in sorted(mod.connections_to, key=lambda c: c.to_input_name):
        step = _add_step(steps, modules, params, module_to_step, modules[conn.from_module_id])

        # index is a special connection to keep the order of steps in fixed pipeline templates
        if 'index' in conn.to_input_name:
            continue

        if step.startswith('inputs.'):
            inputs[conn.to_input_name] = step
        else:
            if conn.to_input_name in inputs:
                previous_value = inputs[conn.to_input_name]
                if isinstance(previous_value, str):
                    inputs[conn.to_input_name] = [previous_value] + ['%s.%s' % (step, conn.from_output_name)]
                else:
                    inputs[conn.to_input_name].append('%s.%s' % (step, conn.from_output_name))
            else:
                inputs[conn.to_input_name] = '%s.%s' % (step, conn.from_output_name)

    outputs = sorted(set([c.from_output_name for c in mod.connections_from if c.from_output_name != 'index']))
    if len(outputs) == 0:  # Add 'produce' output for the last step of the pipeline
        outputs = {'produce'}

    klass = get_class(mod.name)
    primitive_desc = {
        key: value
        for key, value in klass.metadata.query().items()
        if key in {'id', 'version', 'python_path', 'name', 'digest'}
    }

    # Create step description
    if len(inputs) > 0:
        step = {
            'type': 'PRIMITIVE',
            'primitive': primitive_desc,
            'arguments': {
                name: {
                    'type': 'CONTAINER',
                    'data': data,
                }
                for name, data in inputs.items()
            },
            'outputs': [{'id': o} for o in outputs]
        }
    else:
        step = {
            'type': 'PRIMITIVE',
            'primitive': primitive_desc,
        }

    # If hyperparameters are set, export them
    if mod.id in params and 'hyperparams' in params[mod.id]:
        hyperparams = pickle.loads(params[mod.id]['hyperparams'])
        # We check whether the hyperparameters have a value or the complete description
        hyperparams = {
            k: {'type': v['type'] if isinstance(v, dict) and 'type' in v else 'VALUE',
                'data': v['data'] if isinstance(v, dict) and 'data' in v else v}
            for k, v in hyperparams.items()
        }
        step['hyperparams'] = hyperparams

    step_nb = 'steps.%d' % len(steps)
    steps.append(step)
    module_to_step[mod.id] = step_nb

    return step_nb


def to_d3m_json(pipeline):
    """Converts a Pipeline to the JSON schema from metalearning working group.
    """
    steps = []
    modules = {mod.id: mod for mod in pipeline.modules}
    params = {}
    for param in pipeline.parameters:
        params.setdefault(param.module_id, {})[param.name] = param.value
    module_to_step = {}
    for _, mod in sorted(modules.items(), key=lambda x: x[0]):
        _add_step(steps, modules, params, module_to_step, mod)

    json_pipeline = {
        'id': str(pipeline.id),
        'name': str(pipeline.id),
        'description': pipeline.origin or '',
        'schema': 'https://metadata.datadrivendiscovery.org/schemas/'
                  'v0/pipeline.json',
        'created': pipeline.created_date.isoformat() + 'Z',
        'context': 'TESTING',
        'inputs': [
            {'name': "input dataset"},
        ],
        'outputs': [
            {
                'data': 'steps.%d.produce' % (len(steps) - 1),
                'name': "predictions",
            }
        ],
        'steps': steps,
    }

    try:
        json_pipeline['digest'] = compute_digest(json_pipeline)
    except:
        logger.error('Creating digest, skipping this field')

    return json_pipeline



import io
import os
import setuptools


# pip workaround
os.chdir(os.path.abspath(os.path.dirname(__file__)))


with io.open('README.md', encoding='utf-8') as fp:
    description = fp.read()

req = [
    # Non-D3M dependencies:
    'Cython',
    'SQLAlchemy==1.2.16',
    'datamart-materialize>=0.9,<0.10',
    'datamart-profiler>=0.9,<0.10',
    'nltk',
    'numpy',
    'scipy',
    'smac>=0.13,<0.14',
    'ConfigSpace>=0.4.20,<0.5',
    'scikit-learn==0.22.2.post1',
    'torch>=1.7',
    'PyYAML',

    # D3M dependencies:
    'd3m==2021.12.19',
    'd3m-automl-rpc==1.2.0',
    'metalearn==0.6.2',
]

setuptools.setup(name='alphad3m',
      version='0.11.0',
      packages=setuptools.find_packages(),
      entry_points={
          'console_scripts': [
              'alphad3m_serve = alphad3m.main:main_serve',
              'alphad3m_search = alphad3m.main:main_search'
              ]},
      install_requires=req,
      description="AlphaD3M: NYU's AutoML System",
      long_description=description,
      long_description_content_type='text/markdown',
      include_package_data=True,
      author='Remi Rampin, Roque Lopez, Raoni Lourenco',
      author_email='remi.rampin@nyu.edu, rlopez@nyu.edu, raoni@nyu.edu',
      maintainer='Remi Rampin, Roque Lopez, Raoni Lourenco',
      maintainer_email='remi.rampin@nyu.edu, rlopez@nyu.edu, raoni@nyu.edu',
      keywords=['datadrivendiscovery', 'automl', 'd3m', 'ta2', 'nyu'],
      license='Apache-2.0',
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: Apache Software License',
          'Topic :: Scientific/Engineering',
      ])



import logging

import torch
import torch.nn as nn
import torch.nn.functional as F

logger = logging.getLogger(__name__)


class PipelineNNet(nn.Module):
    def __init__(self, game, args):
        # game params
        self.action_size = game.getActionSize()
        self.board_size = game.getBoardSize()
        self.problem = game.problem
        self.args = args

        super(PipelineNNet, self).__init__()
        hlayer = 512
        torch.manual_seed(1)
        self.lstm = nn.LSTM(self.board_size, hlayer, 2)
        self.probFC = nn.Linear(hlayer, self.action_size)
        self.valueFC = nn.Linear(hlayer, 1)

    def forward(self, s):
        s = s.view(-1, 1, self.board_size)
        lstm_out, hidden = self.lstm(s)
        s = lstm_out[:,-1]
        pi = self.probFC(s)                                                                         # batch_size x 512
        v = self.valueFC(s)                                                                          # batch_size x 512
                
        if self.problem == 'CLASSIFICATION':
            return F.log_softmax(pi, 1), F.sigmoid(v)
        else:
            return F.log_softmax(pi, 1), v



import logging
import time
import numpy as np
from collections import deque
from alphad3m.pipeline_search.Arena import Arena
from alphad3m.pipeline_search.MCTS import MCTS

#from alphad3m.pipeline_search.utils import Bar, AverageMeter


logger = logging.getLogger(__name__)

np.random.seed(0)


class Coach():
    """
    This class executes the self-play + learning. It uses the functions defined
    in Game and NeuralNet. args are specified in main.py.
    """
    def __init__(self, game, nnet, args):
        self.game = game
        self.board = game.getInitBoard()
        self.nnet = nnet
        self.args = args
        self.mcts = MCTS(self.game, self.nnet, self.args)

    def executeEpisode(self):
        """
        This function executes one episode of self-play, starting with player 1.
        As the game is played, each turn is added as a training example to
        trainExamples. The game is played till the game ends. After the game
        ends, the outcome of the game is used to assign values to each example
        in trainExamples.

        It uses a temp=1 if episodeStep < tempThreshold, and thereafter
        uses temp=0.

        Returns:
            trainExamples: a list of examples of the form (canonicalBoard,pi,v)
                           pi is the MCTS informed policy vector, v is +1 if
                           the player eventually won the game, else -1.
        """
        trainExamples = []
        self.board = self.game.getInitBoard()
        self.curPlayer = 1
        episodeStep = 0
        
        while True:
            episodeStep += 1
            canonicalBoard = self.game.getCanonicalForm(self.board, self.curPlayer)
            temp = int(episodeStep < self.args.get('tempThreshold'))

            pi = self.mcts.getActionProb(canonicalBoard, temp=temp)

            sym = self.game.getTrainExamples(canonicalBoard, pi)

            trainExamples.append(sym)

            action = np.random.choice(len(pi), p=pi)

            logger.info('COACH ACTION %s', action)
            self.board, self.curPlayer = self.game.getNextState(self.board, self.curPlayer, action)

            r = self.game.getGameEnded(self.board, self.curPlayer)
            if r != 0:
                break

        return trainExamples
    
    def learn(self):
        """
        Performs numIters iterations with numEps episodes of self-play in each
        iteration. After every iteration, it retrains neural network with
        examples in trainExamples (which has a maximium length of maxlenofQueue).
        It then pits the new neural network against the old one and accepts it
        only if it wins >= updateThreshold fraction of games.
        """

        trainExamples = deque([], maxlen=self.args.get('maxlenOfQueue'))
        for i in range(self.args.get('numIters')):
            # bookkeeping
            logger.info('------ITER ' + str(i+1) + '------')
            #eps_time = AverageMeter()
            #bar = Bar('Self Play', max=self.args.get('numEps'))
            end = time.time()

            for eps in range(self.args.get('numEps')):
                self.mcts = MCTS(self.game, self.nnet, self.args)   # reset search tree
                trainExamples += self.executeEpisode()

                # bookkeeping + plot progress
                #eps_time.update(time.time() - end)
                end = time.time()
                #bar.suffix = '({eps}/{maxeps}) Eps Time: {et:.3f}s | Total: {total:} | ETA: {eta:}'.format(eps=eps+1, maxeps=self.args.get('numEps'), et=eps_time.avg,
                #                                                                                           total=bar.elapsed_td, eta=bar.eta_td)
                #bar.next()
            #bar.finish()
            
            # training new network, keeping a copy of the old one
            self.nnet.save_checkpoint(folder=self.args.get('checkpoint'), filename='temp.pth.tar')
            pnet = self.nnet.__class__(self.game)
            pnet.load_checkpoint(folder=self.args.get('checkpoint'), filename='temp.pth.tar')
            pmcts = MCTS(self.game, pnet, self.args)
            boards, pis, vs = list(zip(*trainExamples))
            #logger.info([board[self.game.m:self.game.m+self.game.p] for board in boards])
            self.nnet.train(trainExamples)
            nmcts = MCTS(self.game, self.nnet, self.args)

            logger.info('PITTING AGAINST PREVIOUS VERSION')
            arena = Arena(lambda x: np.argmax(pmcts.getActionProb(x, temp=0)),
                          lambda x: np.argmax(nmcts.getActionProb(x, temp=0)), self.game, self.game.display, self.args['stepsfile'])
            pwins, nwins = arena.playGames(self.args.get('arenaCompare'), verbose=self.args['verbose'])

            logger.info('EVALUATIONS ', self.game.evaluations)
            logger.info('NEW/PREV WINS : ' + str(nwins) + '/' + str(pwins))
            if float(nwins)/(pwins+nwins) < self.args['updateThreshold']:
                logger.info('REJECTING NEW MODEL')
                self.nnet = pnet

            else:
                logger.info('ACCEPTING NEW MODEL')
                self.nnet.save_checkpoint(folder=self.args['checkpoint'], filename='checkpoint_' + str(i) + '.pth.tar')
                self.nnet.save_checkpoint(folder=self.args['checkpoint'], filename='best.pth.tar')
